/*
   pikt.c -- PIKT script interpreter

   $Header: /home/cvs/ucgsb/pikt/src/piktd/pikt.c,v 1.46 2000/06/04 00:52:20 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2000 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA

   --------------------------------------------------------------------------

   Usage:  pikt [-cGhT] [+M "<mailcmd>"] [+L "<lpcmd>"] [+A|S <alert/script>]

           -c              syntax check alarm script file
           -G              run in debug mode
           -h              show program help
           -T              run in test mode
           +M              mail command
           +L              print command
           +A|S            alert/script

 */

#include "../lib/common.h"
#include "piktd.h"

const char pikt_version[] = "PIKT pikt version " VERSION;

/* command-line options */
int ischeck = FALSE;
int isdebug = FALSE;
int isdoublecheck = FALSE;
int ishelp = FALSE;
int istest = FALSE;

int bailout = FALSE;

int interactive = FALSE;

int prescan;

/* PIKT.conf options */
int uid = 0;		/* default to root */
char *master = (char *) NULL;
char *domain = (char *) NULL;
char *accessmode = (char *) NULL;
char *master_address = (char *) NULL;
int isauth_by_master = TRUE;
int isauth_by_master_address = FALSE;
int rpc_timeout = RPC_TIMEOUT;
int rpc_retry_timeout = RPC_RETRY_TIMEOUT;
int rpcinfo_retries = RPCINFO_RETRIES;
int isverboselog = FALSE;
int isautosyslog = FALSE;
int istcponly = FALSE;
int execcmdout_limit = EXECCMDOUT_LIMIT;

char hostname[MAXIDENT+1];
char domainname[MAXIDENT+1];

char **argvcurr;
char **argvlast;

unsigned stroffset = 0;

int lineno = 0;

char *altcurr;

char *altlockfile;
char *execlockfile;
char *hstfile;

int dohistory = FALSE;

char *piktlog;
/* char syslog[] = "/var/log/syslog"; */

signed long arndx[MAXARYDIM+1];
signed long arndxnil[MAXARYDIM+1] = { NILARNDX, NILARNDX, NILARNDX, NILARNDX };
signed long arndxnul[MAXARYDIM+1] = {       0L,       0L,       0L,       0L };

const char *arytyp[] = { "ARYSTR", "ARYNUM" };	/* string exprinput prefixes
						   as a placeholder for an
						   array
						 */

const signed char err[]    = { ERROR, '\0' };
const signed char nil[]    = { NIL, '\0' };
const char errstr[] = "<<ERR>>";
const char nilstr[] = "<<NIL>>";

const signed char dquote[] = { DQTPROXY, '\0' };
const signed char squote[] = { SQTPROXY, '\0' };
const char space[]  = " ";

int exprtyp;

int exprres;
double exprval;
char *exprstr;

FILE *inputfp = (FILE *) NULL;
ino_t inputino = 0;
off_t inputsize = 0;
unsigned long inputposend = 0;
pid_t inputpid;

FILE *altin;

char *mailcmd, *lpcmd;

alrm *alarms = (alrm *) NULL;

alrm *alarmcurr;
val *valcurr;
ky *scvarcurr;
ky *keycurr;
rul *rulcurr;
tok *tokcurr;
stm *stmcurr;

val **values[2][MAXVARS+1];

val *valp[MAXVAP];
val *valnow = (val *) NULL;
val *valprc = (val *) NULL;

int vapndx = 0;

unsigned long services = ALL_SERVICES;	/* all services enabled by default */

int waitpid_options = 0
#ifdef NEED_WAITPID_WNOHANG
                        | WNOHANG
#endif
                        ;

void loadscripts(char *);
void beginalert(char *);
void endalert(char *);
void runscripts(char *);
int runscript(int, FILE *, int);
int beginscript(void);
int endscript(void);

/* do program initializations; parse the command line; load the alarm scripts;
   run the scripts; send off alert messages, etc., as needed
 */
int
main (int argc, char *argv[])
{
	static char _main_[] = "main";
	pid_t pid;
	static char pidstr[8];

	#if DEEPDEBUG
	char input[MAXLINE];
	char output[MAXLINE];

	while (TRUE) {
		gets(input);
		decompose(input, output);
		printf("%s\n", output);
	}
	#endif

	init(argc, argv, (void (*)(void))pikt_usage);
	if (bailout)
		exit(1);

	argvcurr = argv+1;	/* skip the program name */
	argvlast = argv+argc-1;

	cmdparse();		/* further parse the command line */
	if (ishelp || bailout) {
		pikt_usage();
		exit(1);
	}

	if (! altcurr) {
		/* we haven't yet set piktlog at this point, so we log
		   this to syslog instead; we can safely set isautosyslog
		   to TRUE since we are exiting
		 */
		isautosyslog = TRUE;
		logexit(SYSLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "alert unknown (altcurr is NULL)");
	}
	altlockfile = (char *) mymalloc((strlen(lokdir)+strlen(altcurr)+11) * sizeof(char), _main_);
	sprintf(altlockfile, "%s/%s.alt.lock", lokdir, altcurr);
	execlockfile = (char *) mymalloc((strlen(lokdir)+strlen(altcurr)+12) * sizeof(char), _main_);
	sprintf(execlockfile, "%s/%s.exec.lock", lokdir, altcurr);
	hstfile = (char *) mymalloc((strlen(hstdir)+strlen(altcurr)+6) * sizeof(char), _main_);
	sprintf(hstfile, "%s/%s.hst", hstdir, altcurr);
	/* if logreturn() or logexit() are called before piktlog is set,
	   either should abort
	 */
	piktlog = (char *) mymalloc((strlen(logdir)+strlen(altcurr)+6) * sizeof(char), _main_);
	sprintf(piktlog, "%s/%s.log", logdir, altcurr);

	/* these file checks come later than in the other PIKT binaries,
	   because isfile() can't log any stat() failure until after
	   piktlog is set
	 */
	if (! isfile(RMCMD, TRUE) || ! isfile(TAILCMD, TRUE) || ! isfile(PSCMD, TRUE) || ! isfile(SUMCMD, TRUE) || ! isfile(CKSUMCMD, TRUE)) {
		/* error(s) logged in PIKTLOG */
		exit(1);
	}

	/* apply the lock, to prevent other concurrent pikt sessions
	   for this alert
	 */
	if (islock(altlockfile)) {
		if (interactive) {
			fprintf(stderr, "%s found!  Another %s pikt run may be in progress.\nTry again later (else rm the file).\n", altlockfile, altcurr);
			fflush(stderr);
			exit(1);
		} else {
			/* logexit() automatically deletes the lock file,
			   so we do a logreturn() and exit() here instead
			 */
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "%s found, another %s pikt run may be in progress, run aborted.", altlockfile, altcurr);
			exit(1);
		}
	} else {
		pid = getpid();
		sprintf(pidstr, "%u", (unsigned)pid);
		if (mklock(altlockfile, pidstr, interactive, TRUE) == ERROR)
			logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in main(), mklock() error for %s", altlockfile);
	}

	/* seed the random number generator */
	srand((unsigned)(pid+time((time_t*)0)));

	loadscripts(altcurr);

	if (! ischeck && ! isdoublecheck) {
		beginalert(altcurr);
		runscripts(altcurr);
		endalert(altcurr);
	}

	if (rmlocks(interactive, TRUE) == ERROR)
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in main(), rmlocks() error");

	exit(0);

}

/* output program usage to screen
 */
void
pikt_usage (void)
{
	printf("\nUsage:  pikt [-cGhT] [+M \"<mailcmd>\"] [+L \"<lpcmd>\"] [+A|S <alert/script>]\n\n");
	printf("\t-c\t\tsyntax check alarm script file\n");
	printf("\t-G\t\trun in debug mode\n");
	printf("\t-h\t\tshow program help\n");
	printf("\t-T\t\trun in test mode\n");
	printf("\t+M\t\tmail command\n");
	printf("\t+L\t\tprint command\n");
	printf("\t+A|S\t\talert/script\n\n");
	return;
}

/* load and parse the alarm script file
 */
void
loadscripts (char *alert)
{
	static char _loadscripts_[] = "loadscripts";
	char *altfile;

	if (isdoublecheck) {
		altfile = (char *) mymalloc((strlen(stgdir)+strlen(alert)+6) * sizeof(char), _loadscripts_);
		sprintf(altfile, "%s/%s.alt", stgdir, alert);
	} else {
		altfile = (char *) mymalloc((strlen(altdir)+strlen(alert)+6) * sizeof(char), _loadscripts_);
		sprintf(altfile, "%s/%s.alt", altdir, alert);
	}
	if ((altin = fopen(altfile, "r")) == NULL)
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in loadscripts(), fopen() error for %s", altfile);
	lineno = 1;
	alarmcurr = (alrm *) NULL;
	prescan = TRUE;
	altparse();
	rewind(altin);
	alarmcurr = (alrm *) NULL;
	prescan = FALSE;
	altparse();
	if (fclose(altin))	/* returns 0 on success */
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in loadscripts(), fclose() error for %s", altfile);
	free(altfile);
	return;
}

/* open the history file, if necessary; do some alert initializations
 */
void
beginalert (char *alert)
{
	struct stat fbuf;

	logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "alert run begin");
	if (dohistory) {
		openhistory(hstfile);
	} else {	/* no current history, so remove any old history file */
		if (stat(hstfile, &fbuf) == 0) {
			if (remove(hstfile))
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in beginalert(), remove() error for %s", hstfile);
		}
	}
	keyseplen = strlen(KEYSEP);
	mailmsg = (msg *) NULL;
	printmsg = (msg *) NULL;
	newmailalert = TRUE;
	newprintalert = TRUE;
	return;
}

/* send alert email messages, if necessary; send printer output, if necessary;
   close the history file, if necessary
 */
void
endalert (char *alert)
{
	sendmsg(mailmsg, mailcmd);
	freemsg(&mailmsg);
	sendmsg(printmsg, lpcmd);
	freemsg(&printmsg);
	if (dohistory)
		closehistory(hstfile);
	free(hstfile);
	logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "alert run end");
	return;
}

/* for each alarm script:  initialize; open script input, if called for;
   begin writing the alarm history, if called for; run the script; close
   script input, if called for; finish writing the alarm history, if called
   for; tidy up
 */
void
runscripts (char *alert)
{

	foreach (alarmcurr, alarms) {
		if (STATUS == INACTIVE)
			continue;
		if (beginscript() == ERROR)
			continue;
		if (openinput(&inputfp, &inputpid, INPUTTYP, INPUTSRC) == ERROR)
			continue;
		if (beginhistory() == ERROR)
			continue;
		switch (runscript(INPUTTYP, inputfp, POSTYP)) {
			case ERROR:
				logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "aborted %s", ALMNAM);
				break;
			case QUIT:
				logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "quit %s", ALMNAM);
				break;
			case 0:
				logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "finished %s", ALMNAM);
				break;
		}
		closeinput(inputfp, inputpid);
		endhistory();
		endscript();
	}
	return;
}

/* run the script begin section; and if no errors (and no quit), run the
   script rules section; and if no errors (and no quit), run the script end
   section
 */
int
runscript (int inputtype, FILE *infile, int postype)
{
	int res;

	if ((res = dobegin()) == 0) {
		if ((res = dorules(inputtype, infile, postype)) == 0) {
			res = doend();
		}
	}
	return res;
}

/* remove any potentially blocking exec lock file; do some initializations
 */
int
beginscript (void)
{

	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "ALARM %s", ALMNAM);
	newmailalarm = TRUE;
	newprintalarm = TRUE;
	/* remove any potentially blocking exec lock file
	if (islock(execlockfile)) {
		if (rmlock(execlockfile, interactive, TRUE) == ERROR)
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in beginscript(), rmlock() error for %s", execlockfile);
	}
	 */
	initvals(&valnow);
	initvals(&valprc);
	#if DEEPDEBUG
	if ((HSTTYP & (FIXEDVAL|FREEVAL)) && ! HSTERR) {
		if (isdebug)
			dumpvals(VALHST);
	}
	#endif
	return 0;
}

/* close any open files; garbage collect all unneeded script values
 */
int
endscript (void)
{
	register int i;
	vl *vlp;

	/* close any open files */
	for (i=1; i<=VRNDXHI; i++) {
		vlp = lookupvl(valnow, i, arndxnul, (char *)NULL, ARYTYP(i));
		if (((DATTYP(i) == FILEHANDLE) || (DATTYP(i) == PROCHANDLE)) && vlp && vlp->fil) {
			if (fclose(vlp->fil))
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in endscript(), %s, fclose() error for file handle %s", ALMNAM, VARNAM(i));
		}
	}
	/* we no longer need any of the vals--whether valnows, valprcs, or
	   valhst, so let's free the memory allocated to them; 
	 */
	freevals(&valnow, FIXEDVAL|FREEVAL|BEGINVAL|ENDVAL);
	freevals(&valprc, FIXEDVAL|FREEVAL|BEGINVAL|ENDVAL);
	freevals(&(VALHST), FIXEDVAL|FREEVAL|BEGINVAL|ENDVAL);
	return 0;
}
